// g++ -o bla bla.cpp -lppl -lgmp -O3

#include <array>

// Library for polyhedra
#include "ppl.hh"
namespace PPL = Parma_Polyhedra_Library;
using namespace Parma_Polyhedra_Library::IO_Operators; 

// Library for interval arithmetic
#include <boost/numeric/interval.hpp>
using namespace boost::numeric;
using namespace interval_lib;

// Fix rounding policies for the transcendental functions
typedef interval<double, policies<save_state<rounded_transc_std<double>>, checking_base<double>>> I;

const double min_eps=0.001; // minimum value for eps
const double min_diff_angle=0.0002; // difference with 2pi tolerated to be considered as a (quasi)corona. Sensitive!
const bool asap=true; // how L is defined for the edge potential: as small as possible (asap) if true, as large as possible (alap) otherwise
const bool verbose=false;
const float prec_delta=0.0001; // precision on the upper bound. decrease to speed up (at least for small r) ?


// interval variables
I r; // radius
I delta;
I V111, V11r, V1r1, V1rr, Vr1r, Vrrr; // base vertex potentials (parameters : V1rr==x, V11r==y)
I E111, E11r, E1rr, Errr; // emptiness

// double variables
double delta_inf,delta_sup;
double m1, mr; // coefficient of angle deviation from base vertex potential
double Z1, Zr; // ceiling of vertex potential
double eps; // for epsilon-tight triangles
double l11,l1r,lrr,q11,q1r,qrr; // for edge potential

// eventual exact coronas -> additional equations on x/y to ensure theoreticaly vertex positivity
bool exists_large_corona;
bool exists_small_corona;
int large_corona[3] = {0,0,0};
int small_corona[3] = {0,0,0};

// variables for the polytopes of parameters X=V1rr, Y=V11r
PPL::Variable X(0),Y(1),M1(2),Mr(3);

// system of constraints
PPL::Constraint_System ieqs;

int nb,nt,ng; // used to count the number of different triangles checked
int type_error; // 0: polytope vide, 1: eps=0, 2: local inequality


// stupid but useful function
I val(char q)
{
    return q=='1'?I(1):r;
}

// another stupid but useful function
I Vijk(char i,char j, char k)
{
    std::string t;
    t+=i;
    t+=j;
    t+=k;    
    if (t=="111") return V111;
    if (t=="11r"||t=="r11") return V11r;
    if (t=="r1r") return Vr1r;
    if (t=="1r1") return V1r1;
    if (t=="1rr"||t=="rr1") return V1rr;
    return Vrrr;
}


#include "triangles.cpp"

#include "parameters_xy.cpp"

#include "parameters_others.cpp"



/**********************************************************************************/
/* find by dichotomy the lowest delta which bounds from above the maximal density */
/**********************************************************************************/

// to speed up :
// the first pass does not check the local inequality.
// it is faster and yields a better delta_inf which can then be used in the second pass
// in the second pass, all is checked (including the local inequality)

void find_delta(bool firstpass=true)
{
    // delta_inf always bounds from below the maximal density
    // initially: slightly less than HCP pi/sqrt(12)
    if (firstpass) delta_inf=0.9068;

    // delta_sup always bounds from above the maximal density
    // initially: Florian's upper bound
    delta_sup=upper((asin(r/(I(1)+r))*(I(1)-square(r))+pi<I>()/I(2)*square(r))/(r*sqrt(I(1)+I(2)*r)));
    
    // while the gap bewteen upper and lower bounds on the maximal density is too large
    while (delta_sup-delta_inf>prec_delta or delta_sup!=upper(delta))
    {
        
        if (verbose) printf(".10f <= delta <= %.10f\n",delta_inf,delta_sup);

        // new candidate bound for the density: the median
        delta=I((delta_inf+delta_sup)/2);
        // we have to decide whether it is an upper bound (->replaces delta_sup) or lower bound (->replaces delta_inf)

        
        // update values of emptiness (which depend on delta)
        E111=E(I(2),I(2),I(2),I(1),I(1),I(1));
        E11r=E(I(2),I(1)+r,I(1)+r,r,I(1),I(1));
        E1rr=E(I(2)*r,I(1)+r,I(1)+r,I(1),r,r);
        Errr=E(I(2)*r,I(2)*r,I(2)*r,r,r,r);

        // build the approximated polytope of allowed (x,y,m1,mr)
        // approximation is forced because real constraints are transcendental whereas accepted ones are rational
        // first clear all inequalities
        ieqs.clear();
        // then add constraints to have eps>=min_eps
        add_epsilon_constraints();
        // finally add constraints to have positivity around any vertex
        exists_large_corona=false;
        exists_small_corona=false;
        add_vertex_positivity_constraints('1');
        add_vertex_positivity_constraints('r');
        // build the polytope
        PPL::C_Polyhedron P(ieqs);

        // empty polytope -> let us search for a larger delta
        if (P.is_empty())
        {
            if (verbose) printf("empty polytope\n");
            delta_inf=lower(delta); // delta is a lower bound -> replaces delta_inf
            continue;
        }

        // set the parameters x and y
        set_xy_generic();
        if (exists_large_corona || exists_small_corona) set_xy_corona();

        
        // compute exactly m1, mr, Z1, Zr
        try
        {
            set_mZ('1');
            set_mZ('r');
        }
        catch (std::string const &err)
        {
            std::cout << err << std::endl;
        }

        // compute exactly eps
        set_eps();
        if (eps<=min_eps)
        {
            if (verbose) printf("epsilon too small\n");
            delta_inf=lower(delta); // delta is a lower bound -> replaces delta_inf
            continue;
        }

        // stop here if first pass
        if (firstpass)
        {
            delta_sup=upper(delta); // delta is a potential upper bound -> replaces delta_sup
            continue;
        }

        // if second pass, check further if delta is really an upper bound (and not only a potential one)
        
        // fix the edge potential parameters and check the local inequality for each type of triangle
        set_LQ('1','1');
        set_LQ('1','r');
        set_LQ('r','r');

        // check the local inequality
        try
        {
            I b=I(0,upper(I(2)*r));
            
            if (verbose) printf("Check local inequality on T111...\n");
            nt=0; nb=0; ng=0;
            check_local_ieq(I(2)+b, I(2)+b, I(2)+b, '1', '1', '1',0);
            if (verbose) printf("%d good-, %d eps- and %d-nonfeasible triangles\n",ng,nt,nb);

            if (verbose) printf("Check local inequality on T11r...\n");
            nt=0; nb=0; ng=0;
            check_local_ieq(I(2)+b, I(1)+r+b, I(1)+r+b, 'r', '1', '1',0);
            if (verbose) printf("%d good-, %d eps- and %d-nonfeasible triangles\n",ng,nt,nb);

            if (verbose) printf("Check local inequality on T1rr...\n");
            nt=0; nb=0; ng=0;
            check_local_ieq(I(2)*r+b, I(1)+r+b, I(1)+r+b, '1', 'r', 'r',0);
            if (verbose) printf("%d good-, %d eps- and %d-nonfeasible triangles\n",ng,nt,nb);

            if (verbose) printf("Check local inequality on Trrr...\n");
            nt=0; nb=0; ng=0;
            check_local_ieq(I(2)*r+b, I(2)*r+b, I(2)*r+b, 'r', 'r', 'r',0);
            if (verbose) printf("%d good-, %d eps- and %d-nonfeasible triangles\n",ng,nt,nb);
        }
        catch (std::string const &err) // idéalement : devrait arriver rarement si paramètres bien choisis
        {
            delta_inf=lower(delta); // delta is a lower bound -> replaces delta_inf
            if (verbose) printf("local ineq. violated\n");
            if (verbose) std::cout << err << std::endl;
            continue;
        }
        
        delta_sup=upper(delta); // delta is an upper bound -> replaces delta_sup
    }
    delta=I(delta_sup); // upper bound on the maximal density
}



/************************************************************************************/
// longueur intervalle : 1e-5 semble un bon compromis.
// faut parfois descendre à 1e-6 près des couronnes
// 1e-4 suffit loin des pics
// 1e-3 trop grand

// upper_bound a b d : de a/d à b/d par pas de 1/d
int main(int argc, char *argv[])
{
    long a=atol(argv[1]);
    long b=atol(argv[2]);
    long d=atol(argv[3]);
    time_t time1=std::time(NULL);
    for(long i=b;i>a;i--)
    {
        // subinterval for ratio r
//        r=I(i,i+1)/I(d);
        r=I(i)/I(d);

//        prec_delta=0.001/(1+100*(median(r)-0.2));

        printf("[%d, ",i);std::cout.flush();

        // first pass
        find_delta(true);
        printf("%.30f, ",lower(delta));std::cout.flush();
        
        // second pass
        find_delta(false);
//        printf("%.30f, %.30f, ",median(V1rr),median(V11r));
        printf("%.30f],\n",lower(delta));
    }
    time_t time2=std::time(NULL);
    double seconds = std::difftime(time2,time1);
    printf("%.f minutes\n",seconds/60);
    
    return 0;
}   
